;; ######################################################################
;;
;; DSSSL style sheet for XML specifications
;; Jon Bosak, Sun Microsystems 1998.02.09
;; Modified by Norman Walsh 1998.05.16
;; HTML table code by Chris Maden
;; Other contributions from Bill Lindsey and, as usual, James Clark
;;
;; $Id: xmlspec.dsl 1.5 1998/05/26 10:13:23 nwalsh Exp $
;;
;; ######################################################################


;; =========================== PARAMETERS ==============================

;; for possible use in adjusting title page spacing

(define %squeeze-max?% #t)
(define %tpage-label-factor% 0.75)
(define %tpage-title-factor% 0.75)
(define %tpage-para-sep% 6pt)
(define %tpage-block-sep% 12pt)

(define %titlepage-graphic% "w3c_home.png")

;; adjust these as needed to fit the TOC on one page

(define %toc-level1-factor% 0.8)
(define %toc-max-level% 3)
(define %toc-backtitle-factor% 1.0)
(define %toc-head-after-factor% 1.0)
(define %toc-block-sep% 12pt)

;; for adjusting vertical space throughout the document

(define %master-vert-space-factor% 1.0)

;; basic vertical space settings

(define %para-sep% (* 6pt %master-vert-space-factor%))
(define %block-sep% (* 12pt %master-vert-space-factor%))
(define %head-before-factor% (* 0.75 %master-vert-space-factor%))
(define %head-after-factor% (* 0.5 %master-vert-space-factor%))
(define %line-spacing-factor% 1.1)

;; these font selections are for Windows 95

(define %title-font-family% "Arial")
(define %body-font-family% "Times New Roman")
(define %mono-font-family% "Courier New")
(define %dingbat-font-family% "WingDings")

(define %text-width% 39pica)
(define %body-start-indent% 3pica)
(define %body-width% (- %text-width% %body-start-indent%))

;; ============================== FUNCTIONS =============================

(define OUTER-PARENT-LIST
  (list "spec" "front" "body" "back"
	"div1" "inform-div1" "div2" "div3" "div4" "div5"))

(define LIST-LIST
  (list "slist" "ulist" "olist" "glist" "blist" "orglist"))

(define (INLIST?)
  (let loop ((rest-of-list LIST-LIST))
       (if (null? rest-of-list)
	   #f
	   (if (have-ancestor? (car rest-of-list))
	       #t
	       (loop (cdr rest-of-list))))))

(define ADMON-LIST
  (list "note" "vcnote" "wfcnote"))

(define (INADMON?)
  (let loop ((rest-of-list ADMON-LIST))
       (if (null? rest-of-list)
	   #f
	   (if (have-ancestor? (car rest-of-list))
	       #t
	       (loop (cdr rest-of-list))))))

(define (SECTLEVEL)
  (cond
   ((have-ancestor? "div5") 5)
   ((have-ancestor? "div4") 4)
   ((have-ancestor? "div3") 3)
   ((have-ancestor? "div2") 2)
   ((have-ancestor? "div1") 1)
   ((have-ancestor? "inform-div1") 1)
   (else 0)))

(define (OLSTEP)
  (case
   (modulo (length (hierarchical-number-recursive "OLIST")) 4)
	((1) 14pt)
	((2) 12pt)
	((3) 16pt)
	((0) 14pt)))

(define (ILSTEP) 1pica)

(define (BODYRULE sfore saft)
   (make rule
	 orientation: 'horizontal
	 space-before: sfore
	 space-after: saft
	 line-thickness: 1pt
	 length: %body-width%
	 start-indent: (if (member (current-node) OUTER-PARENT-LIST)
			   %body-start-indent%
			 (inherited-start-indent))
	 display-alignment: 'start))

;; ============================== UNITS ================================

(define-unit px (/ 1in 96))

;; ========================= COMMON FUNCTIONS ===========================

(define (expt b n)
  (if (= n 0)
      1
      (* b (expt b (- n 1)))))

(define (absolute-child-number snl)
  (+ 1 (node-list-length (preced snl))) )

(define if-front-page
  (external-procedure "UNREGISTERED::James Clark//Procedure::if-front-page"))

(define if-first-page
  (external-procedure "UNREGISTERED::James Clark//Procedure::if-first-page"))

(declare-characteristic page-number-format
   "UNREGISTERED::James Clark//Characteristic::page-number-format" "1")

(declare-characteristic page-number-restart?
   "UNREGISTERED::James Clark//Characteristic::page-number-restart?" #f)

(declare-characteristic page-n-columns
   "UNREGISTERED::James Clark//Characteristic::page-n-columns" 1)

(declare-characteristic page-balance-columns?
   "UNREGISTERED::James Clark//Characteristic::page-balance-columns?" #f)

(declare-characteristic page-column-sep
   "UNREGISTERED::James Clark//Characteristic::page-column-sep" 3pica)

(define (HSIZE n)
  (let ((m (if (< n 0) 0 n)))
    (* 10pt
       (expt 1.2 m))))

(define (TSIZE n)
  (let ((m (if (< n 0) 0 n)))
    (* 8pt
       (expt 1.1 m))))

(define (PARNUM)
  (child-number (parent (current-node))))

(define (FNUM n fmt)
  (if (number? n)
      (format-number n fmt)
      #f))

;; ====================== COMMON STYLE TEMPLATES =======================

(define ($block-container$)
  (make display-group
	space-before: %block-sep%
	space-after: %block-sep%
	start-indent: %body-start-indent%))

(define ($lowtitle$ tlevel)
  (let ((hs (HSIZE (- 3 tlevel))))
    (make paragraph
	  use: title-style
	  font-size: hs
	  line-spacing: (* hs %line-spacing-factor%)
	  space-before: (* hs %head-before-factor%)
	  space-after: (* hs %head-after-factor%)
	  keep-with-next?: #t
	  start-indent: %body-start-indent%)))

(define ($italic-seq$)
  (make sequence
    font-posture: 'italic))

(define ($bold-seq$)
  (make sequence
    font-weight: 'bold))

(define ($mono-seq$)
  (make sequence
	font-family-name: %mono-font-family%))

;; =========================== COMMON STYLES ============================

(define header-style
  (style
   font-size: 10pt
   line-spacing: 10pt
   font-posture: 'italic))

(define para-style
  (style
   font-size: 10pt
   font-weight: 'medium
   font-posture: 'upright
   font-family-name: %body-font-family%
   line-spacing: 11pt))

(define monopara-style
  (style
   font-family-name: %mono-font-family%
   font-size: 9pt
   font-posture: 'upright
   line-spacing: 11pt
   start-indent: 0pt
   quadding: 'start))

(define verbatim-style
  (style
   font-family-name: %mono-font-family%
   font-size: 9pt
   font-weight: 'medium
   font-posture: 'upright
   line-spacing: 11pt
   lines: 'asis
   start-indent: 0pt
   input-whitespace-treatment: 'preserve
   quadding: 'start))

(define title-style
  (style
   font-family-name: %title-font-family%
   font-weight: 'bold
   quadding: 'start))

(define admon-para-style
  (style
   font-size: 8pt
   font-weight: 'medium
   font-posture: 'upright
   font-family-name: %title-font-family%
   line-spacing: 10pt))

;; ========================= TITLE PAGE AND TOC =========================

;; ................................ META ................................
;; printing suppressed to control the order of display

(element title (empty-sosofo))
(element version (empty-sosofo))
(element w3c-designation (empty-sosofo))
(element w3c-doctype (empty-sosofo))
(element pubdate (empty-sosofo))
(element notice (empty-sosofo))
(element publoc (empty-sosofo))
(element latestloc (empty-sosofo))
(element prevlocs (empty-sosofo))
(element authlist (empty-sosofo))
(element status (empty-sosofo))
(element abstract (empty-sosofo))
(element pubstmt (empty-sosofo))
(element sourcedesc (empty-sosofo))
(element langusage (empty-sosofo))
(element revisiondesc (empty-sosofo))

;; ........................ HEADER (TITLE PAGE) .........................

(element header
  (make simple-page-sequence
	top-margin: 6pica
	bottom-margin: 6pica
	left-margin: 6pica
	right-margin: 6pica
	input-whitespace-treatment: 'collapse
	(sosofo-append
	 (MAKEW3CHEAD)
	 (MAKETITLE)
	 (MAKEW3CVERSIONING)
	 (MAKETOC))))

;; ............................ MAKEW3CHEAD .............................

(define (MAKEW3CHEAD)
  (make table
	(make table-column
	      width: 19.5pica)
	(make table-column
	      width: 19.5pica)
	(make table-row
	      (make table-cell
		    column-number: 1
		    (make external-graphic
			  entity-system-id: %titlepage-graphic%
			  display?: #t
			  space-before: 0pica
			  space-after: 0pica
			  display-alignment: 'start))
	      (make table-cell
		    column-number: 2
		    cell-row-alignment: 'start
		    (with-mode w3cslug-mode
			       (process-node-list
				(select-elements
				 (descendants
				  (ancestor "spec")) "w3c-designation")))))))

(mode w3cslug-mode
  (element w3c-designation
	   (make paragraph
		 use: title-style
		 quadding: 'end
		 min-leading: 0pt
		 font-size: (HSIZE 3))))

;; ............................. MAKETITLE ..............................

(define (MAKETITLE)
  (with-mode title-mode
	     (process-node-list
	      (select-elements
	       (descendants (ancestor "spec")) "header"))))

(mode title-mode
  (element header (process-children))
  (element w3c-doctype (empty-sosofo))
  (element pubdate (empty-sosofo))
  (element notice (empty-sosofo))
  (element publoc (empty-sosofo))
  (element prevlocs (empty-sosofo))
  (element authlist (empty-sosofo))
  (element status (empty-sosofo))
  (element abstract (empty-sosofo))
  (element pubstmt (empty-sosofo))
  (element sourcedesc (empty-sosofo))
  (element langusage (empty-sosofo))
  (element revisiondesc (empty-sosofo))
  (element title
    (make paragraph
	  space-before: (+ (* (HSIZE 5) %tpage-title-factor%) 18pt)
	  space-after: (- (* (HSIZE 5) %tpage-title-factor%) 6pt)
	  font-family-name: %title-font-family%
	  font-weight: 'bold
	  font-size: (HSIZE 5)
	  line-spacing: (* (HSIZE 5) %line-spacing-factor%
			   %tpage-title-factor%)
	  quadding: 'center
	  keep-with-next?: #t))
  (element version
    (make paragraph
	  font-family-name: %title-font-family%
	  font-weight: 'bold
	  font-size: (HSIZE 5)
	  space-after: (- (* (HSIZE 5) %tpage-title-factor%) 6pt)
	  line-spacing: (* (HSIZE 5) %line-spacing-factor%
			   %tpage-title-factor%)
	  quadding: 'center
	  keep-with-next?: #t)))

;; ......................... MAKEW3CVERSIONING ..........................

(define (MAKEW3CVERSIONING)
  (with-mode w3cversioning-mode
	     (process-node-list
	      (select-elements
	       (descendants (ancestor "spec")) "header"))))

(mode pubdate-mode
      (element pubdate
	       (make sequence
		     use: title-style
		     font-size: (HSIZE 2)
		     (sosofo-append
		      (literal " ")
		      (process-children))))
      (element day
	       (make sequence
		     (literal
		      (string-append
		       (data (current-node))
		       "-"))))
      (element month
	       (make sequence
		     (literal
		      (string-append
		       (substring (data (current-node)) 0 3)
		       "-"))))
      (element year
	       (make sequence
		     (literal
		      (substring (data (current-node)) 2 4)))))


(mode w3cversioning-mode
      (element header (process-children))
      (element title (empty-sosofo))
      (element version (empty-sosofo))
      (element w3c-designation (empty-sosofo))
      (element pubdate (empty-sosofo))
      (element w3c-doctype
	(make paragraph
	      use: title-style
	      font-size: (HSIZE 2)
	      space-before: (* %tpage-block-sep%
			       3 %tpage-title-factor%)
	      space-after: (+ (* %tpage-block-sep%
				 2 %tpage-title-factor%) 12pt)
	      quadding: 'center
	      (process-children)
	      (with-mode pubdate-mode
			 (process-node-list
			  (select-elements
			   (descendants (ancestor "header")) "pubdate")))))
      (element publoc ($w3c-version-label$ "This version"))
      (element prevlocs ($w3c-version-label$ "Previous versions"))
      (element latestloc ($w3c-version-label$ "Latest version"))

      (element (publoc loc) ($w3c-loc-para$))
      (element (prevlocs loc) ($w3c-loc-para$))
      (element (latestloc loc) ($w3c-loc-para$))
      (element loc (process-children)) ;; currently just inherited properties

      (element authlist
	(make paragraph
	      space-before: %tpage-block-sep%
	      space-after: %tpage-block-sep%
	      start-indent: (if %squeeze-max?% 0pt %body-start-indent%)
	      (sosofo-append
	       (make paragraph
		     use: title-style
		     (literal "Editors"))
	       (process-children))))
      (element author
	(make paragraph
	      use: para-style
	      space-before: (if (= 1 (child-number)) %tpage-para-sep% 0pt)
	      start-indent: (if %squeeze-max?% 0pt
			      (+ (inherited-start-indent) 2pica))
	      (sosofo-append
	       (process-first-descendant "name")
	       (literal (if (node-list-empty?
			     (select-elements (descendants (current-node))
					      "affiliation"))
			    ""
			    ", "))
	       (process-first-descendant "affiliation")
	       (make sequence
		     font-posture: 'italic
		     (sosofo-append
		      (literal " (" )
		      (process-first-descendant "email"))
		     (literal ")" ))
	       (let ((part (attribute-string "part")))
		 (if part
		     (literal " [" part "]")
		     (empty-sosofo))))))

      (element status ($w3c-text-module$ "Status of this document"))
      (element (status p) ($w3c-text-module-para$))
      (element (status statusp) ($w3c-text-module-para$))
      (element abstract ($w3c-text-module$ "Abstract"))
      (element (abstract p) ($w3c-text-module-para$))
) ;; end mode

(define ($w3c-version-label$ label-string) ;; used in w3cversioning-mode
  (make paragraph
	use: title-style
	space-before: (* %tpage-block-sep% %tpage-label-factor%)
	space-after: (* %tpage-block-sep% %tpage-label-factor%)
	start-indent: (if %squeeze-max?% 0pt %body-start-indent%)
	(literal label-string)
	(process-children)))

(define ($w3c-text-module$ label-string) ;; used in w3cversioning-mode
  (make paragraph
	space-before: (* %tpage-block-sep% %tpage-label-factor%)
	space-after: (* %tpage-block-sep% %tpage-label-factor%)
	start-indent: (if %squeeze-max?% 0pt %body-start-indent%)
	(sosofo-append
	 (make paragraph
	       use: title-style
	       (literal label-string))
	 (process-children))))

(define ($w3c-text-module-para$) ;; used in w3cversioning-mode
  (make paragraph
	use: para-style
	space-before: (* %tpage-para-sep% %tpage-label-factor%)
	space-after: (* %tpage-para-sep% %tpage-label-factor%)
	start-indent: (if %squeeze-max?% 0pt
			(+ (inherited-start-indent) 2pica))
	end-indent: (if %squeeze-max?% 0pt (+ %body-start-indent% 2pica))
	(process-children)))

(define ($w3c-loc-para$) ;; used in w3cversioning-mode
  (make paragraph
	use: monopara-style
	font-weight: 'medium
	space-before: (if (= 1 (child-number)) %tpage-para-sep% 0pt)
	start-indent: (if %squeeze-max?% 0pt (+ %body-start-indent% 2pica))))
      
;; .............................. MAKETOC ...............................

(define (BLANKPAGE)
   (make paragraph
	 break-before: 'page
	 space-before: 26pica
	 quadding: 'center
	 (literal "This page intentionally left blank.")
	 (empty-sosofo)))

(define (MAKETOC)
  (sosofo-append
   (BLANKPAGE)
   (make paragraph
	 font-family-name: %title-font-family%
	 font-weight: 'bold
	 font-posture: 'upright
	 font-size: (HSIZE 3)
	 line-spacing: (* (HSIZE 3) %line-spacing-factor%)
	 space-after: (* (HSIZE 5) %toc-head-after-factor%)
	 start-indent: 0pt
	 quadding: 'start
	 keep-with-next?: #t
	 break-before: 'page
	 (literal "Table of Contents"))
   (with-mode toc
	      (sosofo-append
	       (process-node-list
		(select-elements
		 (descendants (ancestor "spec")) "body"))
	       (process-node-list
		(select-elements
		 (descendants (ancestor "spec")) "back"))))
   (BLANKPAGE)))

(mode toc
      (element (div1 head)
	       (if (> %toc-max-level% 0) ($toc-entry$ 1) (empty-sosofo)))
      (element (inform-div1 head)
	       (if (> %toc-max-level% 0) ($toc-entry$ 1) (empty-sosofo)))
      (element (div2 head)
	       (if (> %toc-max-level% 1) ($toc-entry$ 2) (empty-sosofo)))
      (element (div3 head)
	       (if (> %toc-max-level% 2) ($toc-entry$ 3) (empty-sosofo)))
      (element (div4 head)
	       (if (> %toc-max-level% 3) ($toc-entry$ 4) (empty-sosofo)))
      (element (div5 head)
	       (if (> %toc-max-level% 4) ($toc-entry$ 5) (empty-sosofo)))
      (default (apply process-matching-children
		      (list "div1" "inform-div1"
			    "div2" "div3" "div4" "div5" "head"))))

(define %toc-indent% 2pica)

(define ($toc-entry$ level)
  (sosofo-append
   (if (and (= level 1)
	    (have-ancestor? "div1")
	    (= (ancestor-child-number "div1") 1)
	    (have-ancestor? "back"))
       (make paragraph
	     use: title-style
	     font-size: (HSIZE 2)
	     start-indent: %body-start-indent%
	     space-before: (* %toc-backtitle-factor% (* %toc-block-sep% 1.5))
	     space-after: (* %toc-backtitle-factor% %toc-block-sep%)
	     (literal "Appendices"))
     (empty-sosofo))
   (make paragraph
	 use: para-style
	 start-indent: (+ %body-start-indent%
			  (* %toc-indent% level))
	 first-line-start-indent: (* -1 %toc-indent%)
	 font-weight: (if (= level 1) 'bold 'medium)
	 space-before: (if (= level 1) (* %toc-level1-factor% 6pt) 0pt)
	 space-after: (if (= level 1) (* %toc-level1-factor% 6pt) 0pt)
	 quadding: 'start
	 (literal
	  (cond
	   ((have-ancestor? "body") (NUMLABEL "1"))
	   ((have-ancestor? "back") (NUMLABEL "A"))
	   (else "")))
	 (make link
	       destination: (current-node-address)
	       (with-mode #f 
			  (sosofo-append
			   (process-children)
			   (literal
			    (if (have-ancestor? "inform-div1")
				" (Non-Normative)" "")))))
	 (make leader (literal "."))
	 (current-node-page-number-sosofo))))

;; ========================== MAJOR DIVISIONS ===========================

(define ($paginated-component$)
  (let ((page-header
	 (make sequence
	       use: header-style
	       (literal
		(data
		 (node-list-first
		  (select-elements 
		   (descendants 
		    (ancestor "spec")) "title"))))))
	(page-footer
	 (make sequence
	       use: header-style
	       (literal "Page ")
	       (page-number-sosofo))))
    (make simple-page-sequence
	  page-number-restart?: (if (equal? (gi) "body") #t #f)
	  page-column-sep: 1pica
	  font-family-name: %body-font-family%
	  font-size: 10pt
	  line-spacing: 11pt
	  start-indent: %body-start-indent%
	  left-header:  (if-front-page (empty-sosofo) page-header)
	  right-header: (if-front-page page-header (empty-sosofo))
	  left-footer: 	(if-front-page (empty-sosofo) page-footer)
	  right-footer: (if-front-page page-footer (empty-sosofo))
	  top-margin: 6pica
	  bottom-margin: 6pica
	  left-margin: 6pica
	  right-margin: 6pica
	  header-margin: 4pica
	  footer-margin: 3pica
	  page-width: 51pica
	  page-height: 66pica
	  input-whitespace-treatment: 'collapse
	  quadding: 'start
	  (make paragraph
		font-family-name: %title-font-family%
		font-weight: 'bold
		font-size: (HSIZE 4)
		line-spacing: (* (HSIZE 4) %line-spacing-factor%)
		space-after: (* (HSIZE 4) %head-after-factor%)
		start-indent: 0pt
		quadding: 'start
		keep-with-next?: #t
		(literal
		 (if (equal? (gi) "back") "Appendices" "")))
	  (process-children))))

(element front ($paginated-component$))
(element body ($paginated-component$))
(element back ($paginated-component$))

;; ============================== SECTIONS ==============================

(define ($section$) ($block-container$))

(define ($sectitle$)
  (let* ((hlevel (SECTLEVEL))
	 (hs (HSIZE (- 4 hlevel))))
    (make paragraph
	  font-family-name: %title-font-family%
	  font-weight:  (if (< hlevel 5) 'bold 'medium)
	  font-posture: (if (< hlevel 5) 'upright 'italic)
	  font-size: hs
	  line-spacing: (* hs %line-spacing-factor%)
	  space-before: (* hs %head-before-factor%)
	  space-after: (* hs %head-after-factor%)
	  first-line-start-indent:
	  (if (< hlevel 3)
	      (* -1 %body-start-indent%)
	    0pt)
	  start-indent: %body-start-indent%
	  quadding: 'start
	  keep-with-next?: #t
	  (sosofo-append
	   (literal
	    (cond
	     ((have-ancestor? "body") (NUMLABEL "1"))
	     ((have-ancestor? "back") (NUMLABEL "A"))
	     (else "")))
	   (process-children)
	   (literal
	    (if (and (= (SECTLEVEL) 1) (have-ancestor? "inform-div1"))
		" (Non-Normative)"
	      ""))))))

(define (NUMLABEL fmt)
  (let ((div1n 
	 (if (have-ancestor? "body")
	     (ancestor-child-number "div1")
	   (if (have-ancestor? "div1")
		 (absolute-child-number (ancestor "div1"))
	     (absolute-child-number (ancestor "inform-div1")))))
	(div2n (ancestor-child-number "div2"))
	(div3n (ancestor-child-number "div3"))
	(div4n (ancestor-child-number "div4"))
	(div5n (ancestor-child-number "div5")))
    (string-append
     (if div1n (string-append (FNUM div1n fmt)) " ")
     (if div2n (string-append "." (FNUM div2n "1"))
       (if div1n ". " " "))
     (if div3n (string-append "." (FNUM div3n "1")) " ")
     (if div4n (string-append "." (FNUM div4n "1")) " ")
     (if div5n (string-append "." (FNUM div5n "1")) " "))))

(define (REFNUMLABEL fmt)
  (let ((div1n 
	 (if (have-ancestor? "body")
	     (ancestor-child-number "div1")
	   (if (have-ancestor? "div1")
		 (absolute-child-number (ancestor "div1"))
	     (absolute-child-number (ancestor "inform-div1")))))
	(div2n (ancestor-child-number "div2"))
	(div3n (ancestor-child-number "div3"))
	(div4n (ancestor-child-number "div4"))
	(div5n (ancestor-child-number "div5")))
    (string-append
     (cond
      ((have-ancestor? "body")
	"Section ")
      ((have-ancestor? "back")
	"Appendix ")
      (else ""))
     (if div1n (string-append (FNUM div1n fmt)) "")
     (if div2n (string-append "." (FNUM div2n "1")) "")
     (if div3n (string-append "." (FNUM div3n "1")) "")
     (if div4n (string-append "." (FNUM div4n "1")) "")
     (if div5n (string-append "." (FNUM div5n "1")) "")
     ": ")))

(element div1 ($section$))
(element (div1 head) ($sectitle$))
(element inform-div1 ($section$))
(element (inform-div1 head) ($sectitle$))
(element div2 ($section$))
(element (div2 head) ($sectitle$))
(element div3 ($section$))
(element (div3 head) ($sectitle$))
(element div4 ($section$))
(element (div4 head) ($sectitle$))
(element div5 ($section$))
(element (div5 head) ($sectitle$))

;; ================= UNCLASSIFIED BLOCK-LEVEL ELEMENTS ==================

(element p
  (make paragraph
	use: para-style
	space-before: %para-sep%
	space-after: %para-sep%
	start-indent: (inherited-start-indent)
	quadding: 'start))

;; titleref in htables
(element (td titleref)
  (let ((cellpadding
	 (let ((cp-attr (inherited-attribute-string "cellpadding"
						    (current-node))))
			      (if cp-attr
				  (* (string->number cp-attr)
				     1pt)
				0pt))))
    (make paragraph
	  start-indent: cellpadding
	  end-indent: cellpadding
	  space-before: cellpadding
	  space-after: cellpadding)))

;; =============================== LISTS ================================

(define (ISPACE spacing)
  (if (equal? "compact" spacing)
      (/ %para-sep% 4)
    (/ %para-sep% 2)))

(define ($list-block$)
  (make display-group
	space-before: (* (if (INLIST?) (/ %para-sep% 2) (/ %block-sep% 2))
			 (if (INADMON?) 0.5 0.8))
	space-after:  (* (if (INLIST?) (/ %para-sep% 2) (/ %block-sep% 2))
			 (if (INADMON?) 0.5 0.8))))

(define ($list-item$)
  (make paragraph
	start-indent: (+ (inherited-start-indent) (ILSTEP))))

(element slist ($list-block$))
;;(element (slist sitem) ($list-item$))
;; nwalsh added bullets 26 May 1998
(element (slist sitem) 
  (make paragraph
    start-indent: (+ (inherited-start-indent) (ILSTEP))
    first-line-start-indent: (- (ILSTEP))
    (make line-field
      field-width: (ILSTEP)
      (literal "\bullet;"))
    (process-children)))

(element ulist ($list-block$))
(element (ulist item) ($list-item$))
(element (ulist item p)
  (let ((spacing (inherited-attribute-string "spacing")))
    (if (= (child-number) 1)
	(let ((ilevel 
	       (length (hierarchical-number-recursive "ulist"))))
	  (make paragraph
	    use: (if (INADMON?) admon-para-style para-style)
	    space-before: (* (ISPACE spacing)
			     (if (INADMON?) 0.5 1.0))
	    first-line-start-indent: (- (ILSTEP))
	    (make line-field
		  font-family-name: %dingbat-font-family%
		  font-size: (if (= ilevel 1)
				 (if (INADMON?) 6pt 7pt)
			       (if (INADMON?) 5pt 6pt))
		  field-width: (ILSTEP)
		  (literal "l"))
	    (make sequence
		  first-line-start-indent: 0pt
		  (process-children))))
      (make paragraph
	    use: (if (INADMON?) admon-para-style para-style)
	    space-before: (* (ISPACE spacing)
			     (if (INADMON?) 0.5 1.0))
	    start-indent: (inherited-start-indent)
	    (process-children)))))

(element olist ($list-block$))
(element (olist item) ($list-item$))
(element (olist item p)
  (let ((spacing (inherited-attribute-string "spacing")))
    (if (= (child-number) 1)
	(make paragraph
	      use: (if (INADMON?) admon-para-style para-style)
	      space-before: (* (ISPACE spacing)
			       (if (INADMON?) 0.5 1.0))
	      start-indent: (+ (inherited-start-indent) (OLSTEP))
	      first-line-start-indent: (- (OLSTEP))
	      (make line-field
		    field-width: (OLSTEP)
		    (literal
		     (case
			 (modulo
			  (length
			   (hierarchical-number-recursive
			    "olist")) 4)
		       ((1) (string-append
			     (format-number
			      (PARNUM) "1") "."))
		       ((2) (string-append 
			     (format-number
			      (PARNUM) "a") "."))
		       ((3) (string-append
			     "(" (format-number
				  (PARNUM) "i") ")"))
		       ((0) (string-append
			     "(" (format-number
				  (PARNUM) "a") ")")))))
	      (make sequence
		    first-line-start-indent: 0pt
		    (process-children)))
      (make paragraph
	    use: (if (INADMON?) admon-para-style para-style)
	    space-before: (* (ISPACE spacing)
			     (if (INADMON?) 0.5 1.0))
	    start-indent: (+ (inherited-start-indent) (OLSTEP))
	    (process-children)))))

(element glist ($list-block$))
(element (glist gitem) ($list-item$))
(element (glist gitem label)
  (make paragraph
	use: para-style
	font-weight: 'bold
	space-before: (* %block-sep% 0.75)
	keep-with-next?: #t))
(element (glist gitem def)
  (make paragraph
	use: para-style
	space-after: (* %block-sep% 0.5)
	start-indent: (+ (inherited-start-indent) (ILSTEP))
	(process-children)))
(element (glist gitem def p)
  (make paragraph
	use: para-style
	space-before: (* %para-sep% 0.5)
	space-after: (* %para-sep% 0.5)
	start-indent: (inherited-start-indent)
	quadding: 'start))

(element blist ($block-container$))
(element bibl
    (sosofo-append
     (make paragraph
	   use: para-style
	   space-before: %para-sep%
	   (literal (attribute-string "key")))
     (make paragraph
	   use: para-style
	   space-after: %para-sep%
	   start-indent: (+ (inherited-start-indent) (ILSTEP))
	   (process-children))))

(element orglist ($list-block$))
(element (orglist member) ($list-item$))
(element (orglist member name)
  (make sequence))
(element (orglist member role)
  (make sequence
	font-posture: 'italic
	(sosofo-append
	 (literal " (")
	 (make sequence
	       (process-children))
	 (literal ")"))))

;; ====================== EXAMPLES AND PRODUCTIONS ======================

(element eg 
  (make paragraph
	space-before: (* %para-sep% 0.8)
	space-after:  %para-sep%
	start-indent: (+ (inherited-start-indent) (ILSTEP))
	use: verbatim-style))

(element scrap
  (sosofo-append
   (with-mode scrap-head-mode
	      (process-matching-children (list "head")))
   (BODYRULE 0pt 3pt)
   (process-children)
   (BODYRULE 3pt (* %block-sep% 0.75))))

(element (scrap head) (empty-sosofo)) ; suppress head in place

(mode scrap-head-mode ; print head out of place
      (element head
	       (make paragraph
		     use: title-style
		     space-before: (* %block-sep% 0.75)
		     space-after: (* %block-sep% 0.25)
		     keep-with-next?: #t)))

(element prodgroup
  (make display-group
	space-before: (if (or
			   (first-sibling? (current-node))
			   (last-sibling? (current-node))) 0pt %block-sep%)
	space-after:  (if (or
			   (first-sibling? (current-node))
			   (last-sibling? (current-node))) 0pt %block-sep%)))

;; definition of prod used to go here, now in a separate entity

;; this routine courtesy jjc
(define (process-prod nl)
  (if (node-list-empty? nl)
      (empty-sosofo)
      (let ((nd (node-list-first nl)))
	(case (gi nd)
	  (("lhs" "rhs")
	   (sosofo-append
	    (process-node-list nd)
	    (process-prod (node-list-rest nl))))
	  ((#f)
	   (process-prod (node-list-rest nl)))
	  (else
	   (let loop ((info '())
		      (nl nl))
	     (if (node-list-empty? nl)
		 (pinfo (reverse info))
		 (let ((nd (node-list-first nl)))
		   (if (equal? (gi nd) "rhs")
		       (sosofo-append
			(pinfo (reverse info))
			(process-prod nl))
		       (loop (cons nd info)
			     (node-list-rest nl)))))))))))
	   
(define (pinfo lst)
  (make table-cell
	column-number: 5
	(let loop ((lst lst))
	  (if (null? lst)
	      (empty-sosofo)
	      (sosofo-append
	       (process-node-list (car lst))
	       (loop (cdr lst)))))))
		 
(element lhs
	 (sosofo-append
	  (make table-cell
		column-number: 1
		font-weight: 'bold
		(make paragraph
		      use: para-style
		      start-indent: 0pt
		      font-weight: 'bold
		      (literal
		       (string-append
			"["
			(number->string (element-number (ancestor "prod")))
			"]"))))
	  (make table-cell
		column-number: 2
		(make paragraph
		      use: monopara-style
		      start-indent: 0pt
		      font-weight: 'bold
		      (process-children)))
	  (make table-cell
		column-number: 3
		(make paragraph
		      start-indent: 0pt
		      (literal "::=")))))

(element rhs
	 (make table-cell
	       starts-row?: (not (first-sibling?))
	       column-number: 4
	       n-columns-spanned: (if (or
				       (equal?
					(gi (my-follow (current-node)))
					"rhs")
				       (absolute-last-sibling?))
				       2 1)
	       (make paragraph
		     use: monopara-style
		     start-indent: 0pt
		     (process-children))))

(element com
  (make paragraph
	use: para-style
	font-posture: 'italic
	start-indent: 0pt
	(sosofo-append
	 (literal "/* ")
	 (process-children)
	 (literal " */"))))

(element vc ($vc-wfc$ "VC"))
(element wfc ($vc-wfc$ "WFC"))

(define ($vc-wfc$ vc-or-wfc)
  (make paragraph
	font-family-name: %title-font-family%
	font-size: 8pt
	start-indent: 1pica
	first-line-start-indent: -1pica
	(sosofo-append
	 (literal
	  (string-append "[" vc-or-wfc ": "))
	 (with-mode constraint-head-mode
		    (process-node-list
		     (select-elements
		      (children
		       (element-with-id
			(attribute-string "def")))
		      "head")))
	 (literal "]"))))

(mode constraint-head-mode
      (element head
	       (make sequence)))

;; ============================ ADMONITIONS =============================

(define ($admon$ admon-string)
  (make paragraph
	space-before: %para-sep%
	space-after: %para-sep%
	use: admon-para-style
	start-indent: (+ (inherited-start-indent) (* (ILSTEP) 2))
	end-indent: (* (ILSTEP) 2)
	(make sequence
	      font-weight: 'bold
	      (sosofo-append
	       (literal admon-string)
	       (with-mode admon-head-mode
			  (process-node-list
			   (select-elements
			    (descendants (current-node)) "head")))))
	(process-children)))

(define ($admon-para$)
  (if (= (child-number) 1)
	(make sequence)
    (make paragraph
	  space-before: %para-sep%
	  space-after: %para-sep%
	  use: admon-para-style
	  start-indent: (inherited-start-indent)
	  end-indent: (* (ILSTEP) 2))))

(mode admon-head-mode
  (element head
	    (make sequence
		  (sosofo-append
		   (process-children)
		   (literal ". ")))))

(element (vcnote head) (empty-sosofo))
(element (wfcnote head) (empty-sosofo))

(element (note p) ($admon-para$))
(element (vcnote p) ($admon-para$))
(element (wfcnote p) ($admon-para$))
(element (issue p) ($admon-para$))

(element (ednote edtext) ($admon-para$))
(element (ednote name) (empty-sosofo))

(element note ($admon$ "NOTE: "))
(element vcnote ($admon$ "VALIDITY CONSTRAINT: "))
(element wfcnote ($admon$ "WELL-FORMEDNESS CONSTRAINT: "))

(element issue
  ($admon$ (string-append "Issue ("
			  (let ((id (attribute-string "id")))
			    (substring id
				       6
				       (string-length id)))
			  "): ")))

(element ednote 
  (let ((name (select-elements (children (current-node)) "name")))
    (if (node-list-empty? name)
	($admon$ "Ed. Note: ")
	;; this is fairly badly broken, but it'll do for today; nwalsh
	($admon$ (string-append "Ed. Note (" (data name) "): ")))))

;; ============================== INLINES ===============================

(element bibref
  (make sequence
	(literal "["
		 (attribute-string "key"
				   (element-with-id (attribute-string "ref")))
		 "]")))

(element emph ($italic-seq$))

(element titleref ($italic-seq$))
(element termref (process-children))
(element xtermref ($italic-seq$))

(element specref
  (with-mode specref-head-mode
	     (process-node-list
	      (select-elements
	       (children
		(element-with-id
		 (attribute-string "ref")))
	       "head"))))

(mode specref-head-mode
  (element head
	   (sosofo-append
	    (make sequence
		  font-weight: 'bold
		  (literal
		   (cond
		    ((have-ancestor? "body")
		      (REFNUMLABEL "1"))
		    ((have-ancestor? "back")
		      (REFNUMLABEL "A"))
		    (else ""))))
	    (make sequence
		  font-weight: 'bold
		  (process-children)))))

(element termdef (process-children))
(element (termdef term) ($italic-seq$))

(element loc (process-children))

;(element code ($mono-seq$))
(element kw ($mono-seq$))

(element quote
  (make sequence
    (literal "\left-single-quotation-mark;")
    (process-children-trim)
    (literal "\right-single-quotation-mark;")))

(element lit
  (make sequence
	(sosofo-append
	 (literal (string #\U-2018))
	 (make sequence
	       font-family-name: %mono-font-family%
	       font-size: 9pt
	       (process-children))
	 (literal (string #\U-2019)))))

(element xnt (process-children))

(element nt
  (make sequence
	font-posture: (if (have-ancestor? "prod")
			  'upright
			'italic)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Code for my-preced and my-follow contributed by Bill Lindsey

;; different from ipreced in that it works on a singleton node list.
;; returns the preceding sibling element or empty node list if
;;  this is the first child.

(define (my-preced snl)
  (let loop ((scanned (empty-node-list))
	     (rest (siblings snl)))
    (cond ((node-list-empty? rest)
	   (empty-node-list))
	  ((node-list=? (node-list-first rest) snl)
	   scanned)
	  (else
	   (loop (node-list-first rest)
		 (node-list-rest rest))))))

(define (my-follow snl)
  (let ((sibs (siblings snl)))
    (let loop ((scanned (node-list-first sibs))
	       (rest (node-list-rest sibs)))
      (cond ((node-list-empty? rest)
	     (empty-node-list))
	    ((node-list=? scanned snl)
	     (node-list-first rest))
	    (else
	     (loop (node-list-first rest)
		   (node-list-rest rest)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; A couple of necessary utilities by Bill Lindsey
;;
;; 1) get the list of sibling elements (assumes origin == parent,
;;                                   in other words we're content)

(define (siblings snl)
  (children (parent snl)))

;; 2) (empty-node-list)
;; this doesn't seem to have been supplied with jade, so we go find
;;  an empty node list at the tail of the "current-node" node-list.

(define (empty-node-list)      
  (let loop ((enl (current-node)))
    (if (node-list-empty? enl)
	enl
	(loop (node-list-rest enl)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; HTML Table Model by Chris Maden
;; This version based on Chris Maden's release of 1997.10.30
;; Changes by Jon Bosak to fit into spec style sheet:
;;   Production rule for P removed
;;   All element names, attribute names, and attribute values lowercased
;;   "table" changed to "htable"
;;   "tbody" changed to "htbody"
;;   Added space-before and space-after to htable
;;   Disabled table-width algorithm; hardwired to width of text column
;;   Background of table cells forced to white
;;   Border of empty cells turned off
;;   Added routines from ISO/IEC 10179:
;;      node-list-reduce
;;      node-list-filter
;;      node-list-property
;;      origin
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; DSSSL stylesheet for HTML 4.0 (97-09-17 draft) tables
;; Created for the W3C XML WG by Christopher R. Maden of O'Reilly and
;;    Associates (crism@oreilly.com).
;; The cellpadding attribute needs to be handled by cell contents.
;; Only tested on LTR content.
;; The COL SPAN attribute is handled correctly by Jade, but ignored by
;;    the RTF back-end.
;; Percentage widths on COLs and COLGROUPs appear to be handled by
;;    Jade, but the RTF back-end produces strange results.
;; WIDTH and HEIGHT set directly on table cells is not supported.

;; (get-column-number) analyzes the previous cells in a table for
;; vertical spans in order to determine the proper location for the
;; current cell.  row-num-array is an array of i integers, where i is
;; the number of cells in a row (as far as is known), and where the
;; (i-1)th element is the number of rows filled in that column so
;; far.  Both rows and columns are counted from 0.

(define (get-column-number)
  ;; Loop over the rows of the table part.
  (let row-loop
    ((rows (children (origin (ancestor "tr"
				       (current-node)))))
     (row-num 0)
     (row-num-array '()))
    ;; Loop over the cells in each column, building the array.
    ;; Returns a two-member list; the first is the column-number of
    ;; (current-node), if it was in the row, or -1 if it wasn't.  The
    ;; second is the row-num-array as modified by the row (or the
    ;; empty list, if (current-node) was found.

    (let ((new-info (let cell-loop
		      ((cells (children (node-list-first rows)))
		       (col-num 0)
		       (col-span 0)
		       (cell-row-array row-num-array))
		      (if (node-list-empty? cells)
			  ;; we finished the row
			  (list -1
				cell-row-array)
			  ;; set some useful values
			  (let ((array-length (length cell-row-array))
				(row-dest (+ row-num
					     (- (string->number (attribute-string "rowspan"
										  (node-list-first cells)))
						1)))
				(colspan (string->number (attribute-string "colspan"
									   (node-list-first cells)))))
			    (if (node-list=? (node-list-first cells)
					     (current-node))
				;; this is us!
				(if (and (> array-length
					    col-num)
					 (>= (list-ref cell-row-array
						       col-num)
					     row-num))
				    ;; but we don't fit here - keep
				    ;; trying
				    (cell-loop cells
					       (+ col-num
						  1)
					       0
					       cell-row-array)
				    ;; this is our home - return the
				    ;; column-number.
				    (list col-num
					  '()))
			        ;; this isn't us yet, so cell-loop
			        ;; again with the adjusted array.
			        (cell-loop (if (or (> col-span
						      1)
						   (and (= col-span
							   0)
							(> colspan
							   1))
						   (and (> array-length
							   col-num)
							(>= (list-ref cell-row-array
								      col-num)
							    row-num)))
					       ;; we're in a span
					       ;; that's not done yet,
					       ;; or we don't fit
					       ;; here, so don't knock
					       ;; off the lead cell
					       cells
					       ;; pass the rest of the
					       ;; cells back to the
					       ;; loop
					       (node-list-rest cells))
					   ;; incr the col-num
					   (+ col-num
					      1)
					   ;; note if we're in a span,
					   ;; and how many columns are
					   ;; left
					   (cond ((> col-span
						     0) (- col-span
							   1))
						 ((and (> colspan
							  1)
						       (or (<= array-length
							       col-num)
							   (< (list-ref cell-row-array
									col-num)
							       row-num))) (- colspan
									     1))
						 (else 0))
					   (if (> col-span
						  0)
					       ;; we're in a span, so
					       ;; place this cell
					       ;; whether or not it
					       ;; overlaps something;
					       ;; the markup was in
					       ;; violation anyway
					       (if (<= array-length
						       col-num)
						   (append cell-row-array
							   (list row-dest))
						   (append (reverse (list-tail (reverse cell-row-array)
									       (- array-length
										  col-num)))
							   (list row-dest)
							   (list-tail cell-row-array
								      (+ col-num
									 1))))
					       ;; we're not in a span,
					       ;; so check the array
					       (if (<= array-length
						       col-num)
						   ;; there wasn't an
						   ;; entry for this
						   ;; column yet
						   (append cell-row-array
							   (list row-dest))
						   ;; do we fit in
						   ;; this column?
						   (if (>= (list-ref cell-row-array
								     col-num)
							   row-num)
						       ;; no, so keep
						       ;; the array
						       ;; unchanged
						       cell-row-array
						       ;; yes, so
						       ;; replace this
						       ;; column's
						       ;; entry with a
						       ;; new number
						       (append (reverse (list-tail (reverse cell-row-array)
										   (- array-length
										      col-num)))
							       (list row-dest)
							       (list-tail cell-row-array
									  (+ col-num
									     1)))))))))))))
      ;; what did the row analysis find?
      (if (= (car new-info)
	     -1)
	  ;; the (current-node) was not in the row, so row-loop again
	  ;; with the rest of the rows.
	  (row-loop (node-list-rest rows)
		    (+ row-num
		       1)
		    (car (cdr new-info)))
	  ;; the (current-node) was found; increment the 0-based index
	  ;; for the column-number
	  (+ (car new-info)
	     1)))))

;; converts two-digit hex string to decimal fraction where 0->0 and
;; 255->1

(define (hex-s-to-dec-frac str)
  (/ (string->number str
		     16)
     255))

;; Analyze the table or colgroup.  Create table-columns for each col
;; element, or for each column implied by a cols or span attribute.

(define (make-columns)
  (let col-loop
    ((colspecs (node-list-filter (lambda (snl)
				   (member (gi snl)
					   '("col"
					     "colgroup")))
				 (children (current-node))))
     (columns (attribute-string (if (equal? (gi (current-node))
					    "colgroup")
				    "span"
				    "cols")
				(current-node)))
     (rules (inherited-attribute-string "rules"))
     (cols-so-far 0))
    (if (node-list-empty? colspecs)
	;; if the colspecs are depleted, does the attribute imply that
	;; we need more table-columns?
	(if (or (not columns)
		(<= (string->number columns)
		    cols-so-far))
	    (empty-sosofo)
	    ;; if so, make them; inherit qualities from colgroup if
	    ;; there is one
	    (sosofo-append (make table-column
				 width: (if (equal? (gi (current-node))
						    "colgroup")
					    (let ((cgwidth (attribute-string "width"
									     (current-node))))
					      (if cgwidth
						  ;; is the width in
						  ;; percentage?
						  (if (equal? (string-ref cgwidth
									  (- (string-length cgwidth)
									     1))
							      #\%)
						      ;; yes -
						      ;; multiply by
						      ;; display-size
						      (* (display-size)
							 (/ (string->number (substring cgwidth
										       0
										       (- (string-length cgwidth)
											  1)))
							    100))
						      ;; no - assume
						      ;; it's pixels
						      (* (string->number cgwidth)
							 1pt))
						  (table-unit 1)))
					    (table-unit 1))
				 cell-before-column-border: (if (and (equal? (gi (current-node))
									     "colgroup")
								     (= cols-so-far
									0)
								     (equal? rules
									     "groups"))
								#t
							        (inherited-cell-before-column-border))
				 cell-after-column-border: (if (and (equal? (gi (current-node))
									    "colgroup")
								    (= cols-so-far
								       (- (string->number columns)
									  1)))
							       #t
							       (inherited-cell-after-column-border))
				 quadding: (if (equal? (gi (current-node))
						       "colgroup")
					       (case (attribute-string "align"
								       (current-node))
						     (("left") 'start)
						     (("center") 'center)
						     (("right"
						       "char") 'end)
						     (("justify") 'justify)
						     (else (inherited-quadding)))
					       (inherited-quadding))
				 cell-row-alignment: (case (attribute-string "valign"
									     (current-node))
							   (("top"
							     "baseline") 'start)
							   (("middle") 'center)
							   (("bottom") 'end)
							   ((#f) (inherited-cell-row-alignment))))
			   (col-loop (empty-node-list)
				     columns
				     rules
				     (+ cols-so-far
					1))))
        (sosofo-append (process-node-list (node-list-first colspecs))
		       (col-loop (node-list-rest colspecs)
				 columns
				 rules
				 (+ cols-so-far
				    1))))))

;; Caption for a table, processed first whether it occurs before or
;; after the table matter.
(mode table-caption
      (element caption
	       (make paragraph
		     font-weight: 'bold
		     quadding: 'center
		     space-before: (inherited-line-spacing)
		     keep-with-next?: #t
		     (with-mode #f
				(process-children-trim)))))

;; Caption already processed in table-caption mode.
(element caption
	 (empty-sosofo))

;; Create a table-column based on col element's inherited attributes.
(element col
	 (let ((rules (inherited-attribute-string "rules")))
	   (make table-column
		 n-columns-spanned: (string->number (attribute-string "span"
								      (current-node)))
		 ;; Only inherit the width from colgroup, not from
		 ;; table.
		 width: (let ((colwidth (let ((width (attribute-string "width"
								       (current-node))))
					  (if width
					      width
					      (attribute-string "width"
								(ancestor "colgroup"
									  (current-node)))))))
			  (if colwidth
			      ;; is the width in percentage?
			      (if (equal? (string-ref colwidth
						      (- (string-length colwidth)
							 1))
					  #\%)
				  ;; yes - multiply by display-size
				  (* (display-size)
				     (/ (string->number (substring colwidth
								   0
								   (- (string-length colwidth)
								      1)))
					100))
				  ;; no - assume it's pixels
				  (* (string->number colwidth)
				     1pt))
			      (table-unit 1)))
		 ;; need rules if rules="groups"
		 cell-before-column-border: (if (and (equal? rules
							     "groups")
						     (first-sibling?))
						#t
					        (inherited-cell-before-column-border))
		 cell-after-column-border: (if (and (equal? rules
							    "groups")
						    (last-sibling?)
						    (or (not (attribute-string "span"
									       (ancestor "colgroup")))
							(equal? (string->number (attribute-string "span"
												  (ancestor "colgroup")))
								(child-number (current-node)))))
					       #t
					       (inherited-cell-after-column-border))
		 ;; ALIGN attribute should only come from self or
		 ;; colgroup.
		 quadding: ((lambda (str)
			      (case str
				    (("left") 'start)
				    (("center") 'center)
				    (("right"
				      "char") 'end)
				    (("justify") 'justify)
				    (else (inherited-quadding))))
			    (let ((align (attribute-string "align"
							   (current-node))))
			      (if align
				  align
			          (attribute-string "align"
						    (ancestor "colgroup"
							      (current-node))))))
		 ;; VALIGN should also only come from colgroup, but no
		 ;; other element uses it so
		 ;; (inherited-attribute-string) is safe.
		 cell-row-alignment: ((lambda (str)
					(case str
					      (("top"
						"baseline") 'start)
					      (("middle") 'center)
					      (("bottom") 'end)
					      ((#f) (inherited-cell-row-alignment))))
				      (inherited-attribute-string "valign"
								  (current-node))))))

;; Group of columns; process its children.
(element colgroup
	 (make-columns))

;; Create the caption, then handle the content.
(element htable
	 (sosofo-append (with-mode table-caption
				   (process-first-descendant "caption"))
			(make table
			      ;; inherit the width from flowtree
			      ;; parents if not specified here.
;; space-before and space-after added by bosak 1998.02.08
			      space-before: (* %block-sep% 2)
			      space-after: (* %block-sep% 1.5)

<!-- the table width algorithm is disabled in the XML spec style sheet (bosak 1997.12.03)
			      table-width: (let ((width (attribute-string "width"
									  (current-node))))
					     (if width
						 ;; is the width in
						 ;; percentage?
						 (if (equal? (string-ref width
									 (- (string-length width)
									    1))
							     #\%)
						     ;; yes - multiply
						     ;; by display-
						     ;; size
						     (* (display-size)
							(/ (string->number (substring width
										      0
										      (- (string-length width)
											 1)))
							   100))
						     ;; no - assume
						     ;; it's pixels
						     (* (string->number width)
							1pt))
					         ;; no width; use
						 ;; available space
					         (display-size)))
-->
                              table-width: 36pica
			      ;; BORDER="0" overrides any other
			      ;; border-creating attributes.
			      before-row-border: (if (equal? (attribute-string "border"
									       (current-node))
							     "0")
						     #f
						     (if (member (attribute-string "frame"
										   (current-node))
								 '("void"
								   "below"
								   "lhs"
								   "rhs"
								   "vsides"))
							 #f
						         #t))
			      after-row-border: (if (equal? (attribute-string "border"
									      (current-node))
							    "0")
						    #f
						    (if (member (attribute-string "frame"
										  (current-node))
								'("void"
								  "above"
								  "lhs"
								  "rhs"
								  "vsides"))
							#f
						        #t))
			      before-column-border: (if (equal? (attribute-string "border"
										  (current-node))
								"0")
							#f
						        (if (member (attribute-string "frame"
										      (current-node))
								    '("void"
								      "above"
								      "below"
								      "hsides"
								      "rhs"))
							    #f
							    #t))
			      after-column-border: (if (equal? (attribute-string "border"
										 (current-node))
							       "0")
						       #f
						       (if (member (attribute-string "frame"
										     (current-node))
								   '("void"
								     "above"
								     "below"
								     "hsides"
								     "lhs"))
							   #f
							   #t))
			      ;; Keep the table with its caption.
			      keep-with-previous?: (if (node-list-empty? (node-list-filter (lambda (snl)
											     (equal? (gi snl)
												     "caption"))
											   (children (current-node))))
						       #f
						       #t)
			      ;; Make a content-map to place the table
			      ;; parts correctly.
			      (make table-part
				    content-map: '((thead header)
						   (tbody #f)
						   (tfoot footer))
				    (make-columns)
				    (process-first-descendant "thead")
				    (process-first-descendant "tfoot")
				    (process-matching-children "htbody")))))

;; This will go in the main port of the table-part flow object.
(element htbody
	 (make sequence
	       label: 'tbody
	       quadding: (case (attribute-string "align"
						 (current-node))
			       (("left") 'start)
			       (("center") 'center)
			       (("right"
				 "char") 'end)
			       (("justify") 'justify)
			       (else (inherited-quadding)))
	       cell-row-alignment: (case (attribute-string "valign"
							   (current-node))
					 (("top"
					   "baseline") 'start)
					 (("middle") 'center)
					 (("bottom") 'end)
					 (else (inherited-cell-row-alignment)))))

;; a table cell.
(element td
	 ;; save some useful values
	 (let ((cellspacing (let ((cs-attr (inherited-attribute-string "cellspacing"
								       (current-node))))
			      (if cs-attr
				  (* (string->number cs-attr)
				     1pt)
				  0pt)))
	       ;; bgcolor forced to white for this print style sheet -- bosak
	       (bgcolor "#ffffff")
               ;; this is the original version:
;	       (bgcolor (attribute-string "bgcolor"
;					  (current-node)))
	       (border (inherited-attribute-string "border"
						   (current-node)))
	       (rules (inherited-attribute-string "rules"
						  (current-node))))
	   (make table-cell
		 column-number: (get-column-number)
		 n-columns-spanned: (string->number (attribute-string "colspan"
								      (current-node)))
		 n-rows-spanned: (string->number (attribute-string "rowspan"
								   (current-node)))
		 cell-before-row-margin: cellspacing
		 cell-after-row-margin: cellspacing
		 cell-before-column-margin: cellspacing
		 cell-after-column-margin: cellspacing
		 cell-row-alignment: (case (inherited-attribute-string "valign"
								       (current-node))
					   (("top"
					     "baseline") 'start)
					   (("middle") 'center)
					   (("bottom") 'end)
					   ((#f) (inherited-cell-row-alignment)))
		 cell-background?: (if bgcolor
				       #t
				       #f)
		 background-color: (if bgcolor
				       (color (color-space "ISO/IEC 10179:1996//Color-Space Family::Device RGB")
					      (hex-s-to-dec-frac (substring bgcolor
									    1
									    3))
					      (hex-s-to-dec-frac (substring bgcolor
									    3
									    5))
					      (hex-s-to-dec-frac (substring bgcolor
									    5
									    7)))
				       #f)
		 ;; border="0" overrides any other ruling information.
		 cell-before-row-border: (if (equal? border
						     "0")
					     #f
					     (case rules
						   (("none"
						     "cols") #f)
						   (("groups") (first-sibling? (ancestor "tr")))
						   (else #t)))
		 cell-after-row-border: (if (equal? border
						    "0")
					    #f
					    (case rules
						  (("none"
						    "cols") #f)
						  (("groups") (last-sibling? (ancestor "tr")))
						  (else #t)))
		 cell-before-column-border: (if (equal? border
							"0")
						#f
					        (case rules
						      (("none"
							"rows") #f)
						      (("groups") (inherited-cell-before-column-border))
						      (else #t)))
		 cell-after-column-border: (if (equal? border
						       "0")
					       #f
					       (case rules
						     (("none"
						       "rows") #f)
						     (("groups") (inherited-cell-after-column-border))
						     (else #t)))
		 quadding: (case (attribute-string "align"
						   (current-node))
			       (("left") 'start)
			       (("center") 'center)
			       (("right"
				 "char") 'end)
			       (("justify") 'justify)
			       (else (inherited-quadding)))
		 (if (node-list-empty? (node-list-filter (lambda (snl)
							   (equal? (node-property 'class-name
										  snl)
								   'element))
							 (children (current-node))))
		     (let ((cellpadding (let ((cp-attr (inherited-attribute-string "cellpadding"
										   (current-node))))
					  (if cp-attr
					      (* (string->number cp-attr)
						 1pt)
					      0pt))))
		       (make paragraph
			     start-indent: cellpadding
			     end-indent: cellpadding
			     space-before: cellpadding
			     space-after: cellpadding
			     (process-children)))
		     (process-children)))))

;; This should go in the footer port of the table-part.
(element tfoot
	 (make sequence
	       label: 'tfoot
	       quadding: (case (attribute-string "align"
						 (current-node))
			       (("left") 'start)
			       (("center") 'center)
			       (("right"
				 "char") 'end)
			       (("justify") 'justify)
			       (else (inherited-quadding)))
	       cell-row-alignment: (case (attribute-string "valign"
							   (current-node))
					 (("top"
					   "baseline") 'start)
					 (("middle") 'center)
					 (("bottom") 'end)
					 (else (inherited-cell-row-alignment)))))

;; Identical to TD except bold.
(element th
	 ;; save some useful values
	 (let ((cellspacing (let ((cs-attr (inherited-attribute-string "cellspacing"
								       (current-node))))
			      (if cs-attr
				  (* (string->number cs-attr)
				     1pt)
				  0pt)))
	       (bgcolor (attribute-string "bgcolor"
					  (current-node)))
	       (border (inherited-attribute-string "border"
						   (current-node)))
	       (rules (inherited-attribute-string "rules"
						  (current-node))))
	   (make table-cell
		 font-weight: 'bold
		 column-number: (get-column-number)
		 n-columns-spanned: (string->number (attribute-string "colspan"
								      (current-node)))
		 n-rows-spanned: (string->number (attribute-string "rowspan"
								   (current-node)))
		 cell-before-row-margin: cellspacing
		 cell-after-row-margin: cellspacing
		 cell-before-column-margin: cellspacing
		 cell-after-column-margin: cellspacing
		 cell-row-alignment: (case (inherited-attribute-string "valign"
								       (current-node))
					   (("top"
					     "baseline") 'start)
					   (("middle") 'center)
					   (("bottom") 'end)
					   ((#f) (inherited-cell-row-alignment)))
		 cell-background?: (if bgcolor
				       #t
				       #f)
		 background-color: (if bgcolor
				       (color (color-space "ISO/IEC 10179:1996//Color-Space Family::Device RGB")
					      (hex-s-to-dec-frac (substring bgcolor
									    1
									    3))
					      (hex-s-to-dec-frac (substring bgcolor
									    3
									    5))
					      (hex-s-to-dec-frac (substring bgcolor
									    5
									    7)))
				       #f)
		 ;; BORDER="0" overrides any other ruling information.
		 cell-before-row-border: (if (equal? border
						     "0")
					     #f
					     (case rules
						   (("none"
						     "cols") #f)
						   (("groups") (first-sibling? (ancestor "tr")))
						   (else #t)))
		 cell-after-row-border: (if (equal? border
						    "0")
					    #f
					    (case rules
						  (("none"
						    "cols") #f)
						  (("groups") (last-sibling? (ancestor "tr")))
						  (else #t)))
		 cell-before-column-border: (if (equal? border
							"0")
						#f
					        (case rules
						      (("none"
							"rows") #f)
						      (("groups") (inherited-cell-before-column-border))
						      (else #t)))
		 cell-after-column-border: (if (equal? border
						       "0")
					       #f
					       (case rules
						     (("none"
						       "rows") #f)
						     (("groups") (inherited-cell-after-column-border))
						     (else #t)))
		 quadding: (case (attribute-string "align"
						   (current-node))
			       (("left") 'start)
			       (("center") 'center)
			       (("right"
				 "char") 'end)
			       (("justify") 'justify)
			       (else (inherited-quadding)))
		 (if (node-list-empty? (node-list-filter (lambda (snl)
							   (equal? (node-property 'class-name
										  snl)
								   'element))
							 (children (current-node))))
		     (let ((cellpadding (let ((cp-attr (inherited-attribute-string "cellpadding"
										   (current-node))))
					  (if cp-attr
					      (* (string->number cp-attr)
						 1pt)
					      0pt))))
		       (make paragraph
			     start-indent: cellpadding
			     end-indent: cellpadding
			     space-before: cellpadding
			     space-after: cellpadding
			     (process-children)))
		     (process-children)))))

;; This goes in the header port of the table-part.
(element thead
	 (make sequence
	       label: 'thead
	       quadding: (case (attribute-string "align"
						 (current-node))
			       (("left") 'start)
			       (("center") 'center)
			       (("right"
				 "char") 'end)
			       (("justify") 'justify)
			       (else (inherited-quadding)))
	       cell-row-alignment: (case (attribute-string "valign"
							   (current-node))
					 (("top"
					   "baseline") 'start)
					 (("middle") 'center)
					 (("bottom") 'end)
					 (else (inherited-cell-row-alignment)))))

;; A humble row.
(element tr
	 (make table-row
	       quadding: (case (attribute-string "align"
						 (current-node))
			       (("left") 'start)
			       (("center") 'center)
			       (("right"
				 "char") 'end)
			       (("justify") 'justify)
			       (else (inherited-quadding)))
	       cell-row-alignment: (case (attribute-string "valign"
							   (current-node))
					 (("top"
					   "baseline") 'start)
					 (("middle") 'center)
					 (("bottom") 'end)
					 (else (inherited-cell-row-alignment)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Routines from ISO/IEC 10179

(define (node-list-reduce nl
			  combine
			  init)
  (if (node-list-empty? nl)
      init
      (node-list-reduce (node-list-rest nl)
			combine
			(combine init
				 (node-list-first nl)))))

(define (node-list-filter proc
			  nl)
  (node-list-reduce nl
		    (lambda (result snl)
		      (if (proc snl)
			  (node-list snl
				     result)
			  result))
		    (empty-node-list)))

(define (node-list-property prop nl)
  (node-list-map (lambda (snl)
                   (node-property prop snl default:
 (empty-node-list)))
                 nl))

(define (origin nl)
  (node-list-property 'origin nl))

;; ######################################################################
;;
;; DSSSL style sheet for the additional elements used
;; for the FO description - as modified by Sharon Adler using DynaTag
;; by Anders Berglund 1998.08.01
;;
;; ######################################################################

;;
;; Formatting object summary list - block level styles
;;

(element foname
  (make paragraph
     use: para-style
     space-before: %para-sep%
     space-after: %para-sep%
     start-indent: (+ (inherited-start-indent) 10pt)
     quadding: 'start))

(element attsummary
  (make paragraph
     use: para-style
     space-before: %para-sep%
     space-after: %para-sep%
     start-indent: (+ (inherited-start-indent) 10pt)
     quadding: 'start))

(element attqual
  (make paragraph
     use: para-style
     space-before: 0pt
     space-after: 0pt
     start-indent: (+ (inherited-start-indent) 72pt)
     quadding: 'start))

;;
;; Additions to the CODE inline element
;;

(element code
  (make sequence
    font-family-name:
     (if (equal? (attribute-string "role") "specifier")
       %body-font-family%
       %mono-font-family%)
    font-posture:
     (if (equal? (attribute-string "role") "attrvalue")
       'italic
       (inherited-font-posture))
    font-weight:
     (if (or (equal? (attribute-string "role") "elemname")
             (equal? (attribute-string "role") "featurename"))
       'bold
       (if (or (equal? (attribute-string "role") "attrname")
               (equal? (attribute-string "role") "attrvalue"))
	   'medium
	   (inherited-font-weight)))
))

