;; ######################################################################
;;
;; DSSSL style sheet for XML specifications: HTML output
;; Norman Walsh, ArborText, 1998.05.16; based on 
;;   Jon Bosak, Sun Microsystems 1998.02.08/09
;;
;; $Id: xmlspech.dsl 1.5 1998/05/26 10:13:17 nwalsh Exp $
;;
;; ######################################################################


;; =========================== PARAMETERS ==============================

(define %titlepage-graphic% "http://www.w3.org/pub/WWW/Icons/WWW/w3c_home")

(define %toc-max-level% 3)

(define %output-filename% "spec.html")
(define %output-pubid% "-//W3C//DTD HTML 4.0 Transitional//EN")

;; ============================== FUNCTIONS =============================

(define OUTER-PARENT-LIST
  (list "spec" "front" "body" "back"
	"div1" "inform-div1" "div2" "div3" "div4" "div5"))

(define LIST-LIST
  (list "slist" "ulist" "olist" "glist" "blist" "orglist"))

(define (INLIST?)
  (let loop ((rest-of-list LIST-LIST))
       (if (null? rest-of-list)
	   #f
	   (if (have-ancestor? (car rest-of-list))
	       #t
	       (loop (cdr rest-of-list))))))

(define ADMON-LIST
  (list "note" "vcnote" "wfcnote"))

(define (INADMON?)
  (let loop ((rest-of-list ADMON-LIST))
       (if (null? rest-of-list)
	   #f
	   (if (have-ancestor? (car rest-of-list))
	       #t
	       (loop (cdr rest-of-list))))))

(define (SECTLEVEL)
  (cond
   ((have-ancestor? "div5") 5)
   ((have-ancestor? "div4") 4)
   ((have-ancestor? "div3") 3)
   ((have-ancestor? "div2") 2)
   ((have-ancestor? "div1") 1)
   ((have-ancestor? "inform-div1") 1)
   (else 0)))

(define (OLSTEP)
  (case
   (modulo (length (hierarchical-number-recursive "OLIST")) 4)
	((1) 14pt)
	((2) 12pt)
	((3) 16pt)
	((0) 14pt)))

(define (ILSTEP) 1pica)

(define (BODYRULE)
   (make empty-element gi: "HR"))

(define (generate-anchor #!optional (nd (current-node)))
  (string-append "AEN" (number->string (all-element-number nd))))

(define (generate-secanchor #!optional (nd (current-node)))
  (or (attribute-string "id" (parent nd))
      (generate-anchor nd)))

(define (href-to target)
  ;; Return the HTML HREF for the given node.
  (let* ((id (if (attribute-string "id" target)
		 (attribute-string "id" target)
		 (generate-anchor target)))
	 (fragid (string-append "#" id)))
    fragid))

;; ============================== UNITS ================================

(define-unit px (/ 1in 96))

;; ========================= COMMON FUNCTIONS ===========================

(define (expt b n)
  (if (= n 0)
      1
      (* b (expt b (- n 1)))))

(define (absolute-child-number snl)
  (+ 1 (node-list-length (preced snl))) )

(define if-front-page
  (external-procedure "UNREGISTERED::James Clark//Procedure::if-front-page"))

(define if-first-page
  (external-procedure "UNREGISTERED::James Clark//Procedure::if-first-page"))

(declare-characteristic page-number-format
   "UNREGISTERED::James Clark//Characteristic::page-number-format" "1")

(declare-characteristic page-number-restart?
   "UNREGISTERED::James Clark//Characteristic::page-number-restart?" #f)

(declare-characteristic page-n-columns
   "UNREGISTERED::James Clark//Characteristic::page-n-columns" 1)

(declare-characteristic page-balance-columns?
   "UNREGISTERED::James Clark//Characteristic::page-balance-columns?" #f)

(declare-characteristic page-column-sep
   "UNREGISTERED::James Clark//Characteristic::page-column-sep" 3pica)

(define (HSIZE n)
  (let ((m (if (< n 0) 0 n)))
    (* 10pt
       (expt 1.2 m))))

(define (TSIZE n)
  (let ((m (if (< n 0) 0 n)))
    (* 8pt
       (expt 1.1 m))))

(define (PARNUM)
  (child-number (parent (current-node))))

(define (FNUM n fmt)
  (if (number? n)
      (format-number n fmt)
      #f))

;; ====================== COMMON STYLE TEMPLATES =======================

(define ($block-container$)
  (make element gi: "DIV"
	attributes: (class-attlist)
	(process-children)))

(define ($lowtitle$ tlevel)
  (let (;(hs (HSIZE (- 3 tlevel)))
	(hs (string-append "H" (number->string tlevel))))
    (make element gi: hs
	  (process-children))))

(define ($italic-seq$)
  (make element gi: "I"
	(process-children)))

(define ($bold-seq$)
  (make element gi: "B"
	(process-children)))

(define ($mono-seq$)
  (make element gi: "TT"
	(process-children)))

;; ========================= TITLE PAGE AND TOC =========================

;; ................................ META ................................
;; printing suppressed to control the order of display

(element title (empty-sosofo))
(element version (empty-sosofo))
(element w3c-designation (empty-sosofo))
(element w3c-doctype (empty-sosofo))
(element pubdate (empty-sosofo))
(element notice (empty-sosofo))
(element publoc (empty-sosofo))
(element latestloc (empty-sosofo))
(element prevlocs (empty-sosofo))
(element authlist (empty-sosofo))
(element status (empty-sosofo))
(element abstract (empty-sosofo))
(element pubstmt (empty-sosofo))
(element sourcedesc (empty-sosofo))
(element langusage (empty-sosofo))
(element revisiondesc (empty-sosofo))

;; ........................ HEADER (TITLE PAGE) .........................

(element header
  (sosofo-append
   (MAKEW3CHEAD)
   (MAKETITLE)
   (MAKEW3CVERSIONING)
   (MAKETOC)))

;; ............................ MAKEW3CHEAD .............................

(define (MAKEW3CHEAD)
  (sosofo-append
   (make empty-element gi: "IMG"
	 attributes: (list (list "SRC" %titlepage-graphic%)
			   (list "ALT" "W3C Logo")
			   (list "BORDER" "0")
			   (list "ALIGN" "left")))
   (with-mode w3cslug-mode
	      (process-node-list
	       (select-elements
		(descendants
		 (ancestor "spec")) "w3c-designation")))
   (make empty-element gi: "BR"
	 attributes: '(("CLEAR" "ALL")))))

(mode w3cslug-mode
  (element w3c-designation
    (make element gi: "H4"
	  attributes: '(("ALIGN" "right"))
	  (process-children))))

;; ............................. MAKETITLE ..............................

(define (MAKETITLE)
  (with-mode title-mode
    (process-node-list
     (select-elements
      (descendants (ancestor "spec")) "header"))))

(mode title-mode
  (element header (process-children))
  (element w3c-doctype (empty-sosofo))
  (element pubdate (empty-sosofo))
  (element notice (empty-sosofo))
  (element publoc (empty-sosofo))
  (element prevlocs (empty-sosofo))
  (element authlist (empty-sosofo))
  (element status (empty-sosofo))
  (element abstract (empty-sosofo))
  (element pubstmt (empty-sosofo))
  (element sourcedesc (empty-sosofo))
  (element langusage (empty-sosofo))
  (element revisiondesc (empty-sosofo))
  (element title
    (make element gi: "H1"
	  attributes: '(("ALIGN" "CENTER"))
	  (process-children)))
  (element version
    (make element gi: "H3"
	  attributes: '(("ALIGN" "CENTER"))
	  (process-children))))

;; ......................... MAKEW3CVERSIONING ..........................

(define (MAKEW3CVERSIONING)
  (with-mode w3cversioning-mode
    (process-node-list
     (select-elements
      (descendants (ancestor "spec")) "header"))))

(mode pubdate-mode
  (element pubdate
     (process-children))
  (element day
     (sosofo-append
      (process-children)
      (literal "-")))
  (element month
     (sosofo-append
      (process-children)
      (literal "-")))
  (element year
     (process-children)))

(mode w3cversioning-mode
      (element header (process-children))
      (element title (empty-sosofo))
      (element version (empty-sosofo))
      (element w3c-designation (empty-sosofo))
      (element pubdate (empty-sosofo))
      (element w3c-doctype
	(make element gi: "H3"
	      attributes: '(("ALIGN" "CENTER"))
	      (process-children)
	      (literal " ")
	      (with-mode pubdate-mode
		(process-node-list
		 (select-elements
		  (descendants (ancestor "header")) "pubdate")))))
      (element publoc ($w3c-version-label$ "This version"))
      (element prevlocs ($w3c-version-label$ "Previous versions"))
      (element latestloc ($w3c-version-label$ "Latest version"))

      (element (publoc loc) ($w3c-loc-para$))
      (element (prevlocs loc) ($w3c-loc-para$))
      (element (latestloc loc) ($w3c-loc-para$))

      (element loc ;; currently just inherited properties
	(make element gi: "A"
	      attributes: (list (list "HREF" (attribute-string "href"))
				'("CLASS" "loc"))
	      (process-children)))

      (element authlist
	(sosofo-append
	 (make element gi: "H4"
	       (literal "Editors"))
	 (make element gi: "P"
	       (process-children))
	 ($w3c-legal-boilerplate$)))

      (element author
	(make element gi: "SPAN"
	      attributes: '(("CLASS" "author"))
	      (sosofo-append
	       (process-first-descendant "name")
	       (literal (if (node-list-empty?
			     (select-elements (descendants (current-node))
					      "affiliation"))
			    ""
			    ", "))
	       (process-first-descendant "affiliation")
	       (make element gi: "I"
		     (sosofo-append
		      (literal " (" )
		      (process-first-descendant "email"))
		     (literal ")" ))
	       (let ((part (attribute-string "part")))
		 (if part
		     (literal " [" part "]")
		     (empty-sosofo)))
	       (make empty-element gi: "BR"))))

      (element name
	(make element gi: "SPAN"
	      attributes: '(("CLASS" "name"))
	      (process-children)))

      (element affiliation
	(make element gi: "SPAN"
	      attributes: '(("CLASS" "affiliation"))
	      (process-children)))

      (element email
	(make element gi: "SPAN"
	      attributes: '(("CLASS" "email"))
	      (process-children)))

      (element status ($w3c-text-module$ "Status of this document"))
      (element (status p) ($w3c-text-module-para$))
      (element (status statusp) ($w3c-text-module-para$))
      (element abstract ($w3c-text-module$ "Abstract"))
      (element (abstract p) ($w3c-text-module-para$))
) ;; end mode

(define ($w3c-version-label$ label-string) ;; used in w3cversioning-mode
  (sosofo-append
   (make element gi: "H4"
	 (literal label-string))
   (make element gi: "P"
	 (process-children))))

(define ($w3c-text-module$ label-string) ;; used in w3cversioning-mode
  (sosofo-append
   (make element gi: "H2"
	 attributes: (class-attlist)
	 (literal label-string))
   (process-children)))

(define ($w3c-text-module-para$) ;; used in w3cversioning-mode
  (make element gi: "P"
	(process-children)))

(define ($w3c-loc-para$) ;; used in w3cversioning-mode
  (make element gi: "A"
	attributes: (list (list "HREF" (attribute-string "href"))
			  (list "CLASS" "loc"))
	(process-children)
	(make empty-element gi: "BR")))
      
;; .............................. MAKETOC ...............................

(define (BLANKPAGE)
  (empty-sosofo))

(define (MAKETOC)
  (sosofo-append
   (BLANKPAGE)
   (make element gi: "H2"
	 attributes: '(("CLASS" "table-of-contents"))
	 (literal "Table of Contents"))
   (make element gi: "DL"
	 attributes: '(("CLASS" "table-of-contents"))
	 (with-mode toc
	   (sosofo-append
	    (process-node-list
	     (select-elements
	      (descendants (ancestor "spec")) "body"))
	    (process-node-list
	     (select-elements
	      (descendants (ancestor "spec")) "back")))))
   (BLANKPAGE)))

(define (toc-process-div subdiv process-subdivs?)
  (sosofo-append
    (process-node-list 
     (node-list-first 
      (select-elements (children (current-node)) "head")))
    (if (or (not process-subdivs?)
	    (node-list-empty? (select-elements (children (current-node)) subdiv)))
	(empty-sosofo)
	(make element gi: "DD"
	  (make element gi: "DL"
		(process-node-list 
		 (select-elements (children (current-node)) subdiv)))))))
  
(mode toc
  (element div1
    (toc-process-div "div2" (>= %toc-max-level% 2)))

  (element inform-div1
    (toc-process-div "div2" (>= %toc-max-level% 2)))
  
  (element (div1 head)
    (if (> %toc-max-level% 0) ($toc-entry$ 1) (empty-sosofo)))

  (element (inform-div1 head)
    (if (> %toc-max-level% 0) ($toc-entry$ 1) (empty-sosofo)))

  (element div2
    (toc-process-div "div3" (>= %toc-max-level% 3)))

  (element (div2 head)
    (if (> %toc-max-level% 1) ($toc-entry$ 2) (empty-sosofo)))

  (element div3
    (toc-process-div "div4" (>= %toc-max-level% 4)))

  (element (div3 head)
    (if (> %toc-max-level% 2) ($toc-entry$ 3) (empty-sosofo)))

  (element div4
    (toc-process-div "div5" (>= %toc-max-level% 5)))

  (element (div4 head)
    (if (> %toc-max-level% 3) ($toc-entry$ 4) (empty-sosofo)))

  (element div5
    (toc-process-div "div6" (>= %toc-max-level% 6)))

  (element (div5 head)
    (if (> %toc-max-level% 4) ($toc-entry$ 5) (empty-sosofo))))

(define %toc-indent% 2pica)

(define ($toc-entry$ level)
  (sosofo-append
   (if (and (= level 1)
	    (have-ancestor? "div1")
	    (= (ancestor-child-number "div1") 1)
	    (have-ancestor? "back"))
       (make sequence
	 ;; HACK ALERT!
	 (make formatting-instruction data: "&#60;/DL>")
	 (make element gi: "H3"
	       attributes: '(("CLASS" "table-of-contents"))
	       (literal "Appendices"))
	 (make formatting-instruction data: "&#60;DL>"))
       (empty-sosofo))
   (make element gi: "DT"
	 (literal
	  (cond
	   ((have-ancestor? "body") (NUMLABEL "1"))
	   ((have-ancestor? "back") (NUMLABEL "A"))
	   (else "")))
	 (make element gi: "A"
	       attributes: (list (list "HREF"
				       (string-append "#"
						      (generate-secanchor))))
	       (with-mode #f 
		 (sosofo-append
		  (process-children))))
	 (literal
	  (if (have-ancestor? "inform-div1")
	      " (Non-Normative)" "")))))

;; ========================== MAJOR DIVISIONS ===========================

(element spec
  (make entity
    system-id: %output-filename%
    (make formatting-instruction
	  data: "<![CDATA[<!-- This file is automatically generated. Do not edit! -->]]>&#13;")
    (make document-type
      name: "HTML"
      public-id: %output-pubid%)
    (make element gi: "HTML"
	  (make element gi: "HEAD"
		(make element gi: "TITLE"
		      (with-mode html-title-mode
			(process-first-descendant "title"))))
	  (make element gi: "BODY" 
		(process-children)))))

(define ($paginated-component$)
  (process-children))

(element front ($paginated-component$))
(element body ($paginated-component$))
(element back ($paginated-component$))

(mode html-title-mode
  (default (process-children)))

;; ============================== SECTIONS ==============================

(define ($section$) ($block-container$))

(define ($sectitle$)
  (let* ((hlevel (SECTLEVEL))
	 (hs (string-append "H" (number->string (+ hlevel 1))))
	 (id (generate-secanchor)))
    (make element gi: hs
	  (make element gi: "A"
		attributes: (list (list "NAME" id))
		(literal
		 (cond
		  ((have-ancestor? "body") (NUMLABEL "1"))
		  ((have-ancestor? "back") (NUMLABEL "A"))
		  (else "")))
		(process-children)
		(literal
		 (if (and (= (SECTLEVEL) 1) (have-ancestor? "inform-div1"))
		     " (Non-Normative)"
		     ""))))))

(define (NUMLABEL fmt)
  (let ((div1n 
	 (if (have-ancestor? "body")
	     (ancestor-child-number "div1")
	   (if (have-ancestor? "div1")
		 (absolute-child-number (ancestor "div1"))
	     (absolute-child-number (ancestor "inform-div1")))))
	(div2n (ancestor-child-number "div2"))
	(div3n (ancestor-child-number "div3"))
	(div4n (ancestor-child-number "div4"))
	(div5n (ancestor-child-number "div5")))
    (string-append
     (if div1n (string-append (FNUM div1n fmt)) " ")
     (if div2n (string-append "." (FNUM div2n "1"))
       (if div1n ". " " "))
     (if div3n (string-append "." (FNUM div3n "1")) " ")
     (if div4n (string-append "." (FNUM div4n "1")) " ")
     (if div5n (string-append "." (FNUM div5n "1")) " "))))

(define (REFNUMLABEL fmt)
  (let ((div1n 
	 (if (have-ancestor? "body")
	     (ancestor-child-number "div1")
	   (if (have-ancestor? "div1")
		 (absolute-child-number (ancestor "div1"))
	     (absolute-child-number (ancestor "inform-div1")))))
	(div2n (ancestor-child-number "div2"))
	(div3n (ancestor-child-number "div3"))
	(div4n (ancestor-child-number "div4"))
	(div5n (ancestor-child-number "div5")))
    (string-append
     (cond
      ((have-ancestor? "body")
	"Section ")
      ((have-ancestor? "back")
	"Appendix ")
      (else ""))
     (if div1n (string-append (FNUM div1n fmt)) "")
     (if div2n (string-append "." (FNUM div2n "1")) "")
     (if div3n (string-append "." (FNUM div3n "1")) "")
     (if div4n (string-append "." (FNUM div4n "1")) "")
     (if div5n (string-append "." (FNUM div5n "1")) "")
     ": ")))

(element div1 ($section$))
(element (div1 head) ($sectitle$))
(element inform-div1 ($section$))
(element (inform-div1 head) ($sectitle$))
(element div2 ($section$))
(element (div2 head) ($sectitle$))
(element div3 ($section$))
(element (div3 head) ($sectitle$))
(element div4 ($section$))
(element (div4 head) ($sectitle$))
(element div5 ($section$))
(element (div5 head) ($sectitle$))

;; ================= UNCLASSIFIED BLOCK-LEVEL ELEMENTS ==================

(element p
  (make element gi: "P"
	(process-children)))

;; =============================== LISTS ================================

;; nwalsh added bullets 26 May 1998
(element slist
  (make element gi: "UL"
	attributes: '(("CLASS" "slist"))
	(process-children)))

(element (slist sitem)
  (make element gi: "LI"
	(process-children)))

(element ulist
  (make element gi: "UL"
	(process-children)))

(element (ulist item)
  (make element gi: "LI"
	(process-children)))

(element olist
  (make element gi: "OL"
	(process-children)))

(element (olist item)
  (make element gi: "LI"
	(process-children)))

(element glist 
  (make element gi: "DL"))

(element (glist gitem) 
  (process-children))

(element (glist gitem label)
  (make element gi: "DT"
	(process-children)))

(element (glist gitem def)
  (make element gi: "DD"
	(process-children)))

(element blist (make element gi: "DL"))

(element bibl
    (sosofo-append
     (make element gi: "DT"
	   (literal (attribute-string "key")))
     (make element gi: "DD"
	   (make element
		 gi: "A"
		 attributes: (list (list "NAME" (attribute-string "id")))
		 (empty-sosofo))
	   (process-children))))

;; ====================== EXAMPLES AND PRODUCTIONS ======================

(element eg 
  (make element gi: "PRE"
	attributes: (if (equal? (attribute-string "role") "error")
			'(("style" "color: red"))
			'())
	(process-children)))

(element scrap
  (make element gi: "DIV"
	attributes: '(("CLASS" "scrap"))
	(with-mode scrap-head-mode
	  (process-matching-children (list "head")))
	(BODYRULE)
	(process-children)
	(BODYRULE)))

(element (scrap head) (empty-sosofo)) ; suppress head in place

(mode scrap-head-mode ; print head out of place
  (element head
    (make element gi: "H4"
	  (process-children))))

(element prodgroup
  (make element gi: "DIV"
	(process-children)))

;; definition of prod used to go here, now in a separate entity

;; this routine courtesy jjc, hacked by nwalsh for HTML output.
;; HTML tables use these percentages: [ 5 ] [ 20 ] [ 5 ] [ 40 ] [ 20 ]

(define (process-prod nl)
  (if (node-list-empty? nl)
      (empty-sosofo)
      (let ((nd (node-list-first nl)))
	(case (gi nd)
	  (("lhs" "rhs")
	   (sosofo-append
	    (process-node-list nd)
	    (process-prod (node-list-rest nl))))
	  ((#f)
	   (process-prod (node-list-rest nl)))
	  (else
	   (let loop ((info '())
		      (nl nl))
	     (if (node-list-empty? nl)
		 (pinfo (reverse info))
		 (let ((nd (node-list-first nl)))
		   (if (equal? (gi nd) "rhs")
		       (sosofo-append
			(pinfo (reverse info))
			(process-prod nl))
		       (loop (cons nd info)
			     (node-list-rest nl)))))))))))
	   
(define (pinfo lst)
  (make element gi: "TD"
	attributes: '(("ALIGN" "LEFT")
		      ("VALIGN" "TOP")
		      ("WIDTH" "20%"))
;;	column-number: 5
	(let loop ((lst lst))
	  (if (null? lst)
	      (empty-sosofo)
	      (sosofo-append
	       (process-node-list (car lst))
	       (loop (cdr lst)))))))

(element lhs
  (let ((id (attribute-string "id" (parent (current-node)))))
    (make sequence
      (make element gi: "TH"
	    attributes: '(("ALIGN" "LEFT")
			  ("VALIGN" "TOP")
			  ("WIDTH" "5%"))
; column-number: 1
	    (make element gi: "A"
		  attributes: (list (list "NAME" id))
		  (literal
		   (string-append
		    "["
		    (number->string (element-number (ancestor "prod")))
		    "]"))))
      (make element gi: "TD"
	    attributes: '(("ALIGN" "LEFT")
			  ("VALIGN" "TOP")
			  ("WIDTH" "20%"))
; column-number: 2
	    (process-children))
     
      (make element gi: "TD"
	    attributes: '(("ALIGN" "LEFT")
			  ("VALIGN" "TOP")
			  ("WIDTH" "5%"))
; column-number: 3
	    (literal "::=")))))

(element rhs
  (let ((colspan (if (or (equal? (gi (my-follow (current-node))) "rhs")
			 (absolute-last-sibling?))
		     "2" 
		     "1")))
    (make sequence
      (if (first-sibling?)
	  (empty-sosofo)
	  (make sequence
	    ;; HACK ALERT
	    (make formatting-instruction data: "&#60;/TR>")
	    (make formatting-instruction data: "&#60;TR>")
	    (make formatting-instruction data: "&#60;TD WIDTH=\"5%\">&#38;nbsp;&#60;/TD>")
	    (make formatting-instruction data: "&#60;TD WIDTH=\"20%\">&#38;nbsp;&#60;/TD>")
	    (make formatting-instruction data: "&#60;TD WIDTH=\"5%\">&#38;nbsp;&#60;/TD>")))
      (make element gi: "TD"
	    attributes: (list (list "ALIGN" "LEFT")
			      (list "VALIGN" "TOP")
			      (list "WIDTH" (if (equal? colspan "2")
						"60%"
						"40%"))
			      (list "COLSPAN" colspan))
	    (process-children)))))

(element com
  (make element gi: "P"
	(sosofo-append
	 (literal "/* ")
	 (process-children)
	 (literal " */"))))

(element vc ($vc-wfc$ "VC"))
(element wfc ($vc-wfc$ "WFC"))

(define ($vc-wfc$ vc-or-wfc)
  (let ((target (element-with-id (attribute-string "def"))))
    (make sequence
      (literal (string-append "[" vc-or-wfc ": "))
      (make element gi: "A"
	    attributes: (list (list "HREF" (href-to target)))
	    (with-mode constraint-head-mode
	      (process-node-list
	       (select-elements
		(children
		 (element-with-id
		  (attribute-string "def")))
		"head"))))
      (literal "]"))))

(mode constraint-head-mode
  (element head
    (make element gi: "SPAN"
	  (process-children))))

;; ============================ ADMONITIONS =============================

(define ($admon$ admon-string)
  (let ((id (attribute-string "id")))
    (make element gi: "DIV"
	  attributes: (class-attlist)
	  (make element gi: "BLOCKQUOTE"
		(make element gi: "B"
		      (if id
			  (make element gi: "A"
				attributes: (list (list "NAME" id))
				(literal admon-string))
			  (literal admon-string))
		      (with-mode admon-head-mode
				 (process-node-list
				  (select-elements
				   (descendants (current-node)) "head"))))
		(process-children)))))

(define ($admon-para$)
  (if (= (child-number) 1)
      (process-children)
      ;; NOTE: This may produce invalid HTML because the para created by
      ;; the $admon$ that started all this will not have ended yet.
      ;; Browsers will do the right thing...
      (make element gi: "P"
	    (process-children))))

(mode admon-head-mode
  (element head
    (make element gi: "SPAN"
	  (sosofo-append
	   (process-children)
	   (literal ". ")))))

(element (vcnote head) (empty-sosofo))
(element (wfcnote head) (empty-sosofo))

(element (note p) ($admon-para$))
(element (vcnote p) ($admon-para$))
(element (wfcnote p) ($admon-para$))
(element (issue p) ($admon-para$))

(element note ($admon$ "NOTE: "))
(element vcnote ($admon$ "VALIDITY CONSTRAINT: "))
(element wfcnote ($admon$ "WELL-FORMEDNESS CONSTRAINT: "))

(element issue
  ($admon$ (string-append "Issue ("
			  (let ((id (attribute-string "id")))
			    (substring id
				       6
				       (string-length id)))
			  "): ")))

(element ednote
  (let ((name   (select-elements (children (current-node)) "name"))
	(date   (select-elements (children (current-node)) "date"))
	(edtext (select-elements (children (current-node)) "edtext")))
    (make element gi: "DIV"
	  attributes: '(("CLASS" "ednote"))
	  (make element gi: "BLOCKQUOTE"
		(make element gi: "P"
		      (make element gi: "B"
			    (literal "Ed. Note")
			    (with-mode ednote-mode
			      (make sequence			
				(process-node-list name)
				(process-node-list date)))
			    (literal ": "))
		      (with-mode ednote-mode
			(process-node-list edtext)))))))
  
(mode ednote-mode
  (element name
    (make sequence
      (literal " (")
      (process-children)
      (literal ")")))

  (element date
    (make sequence
      (literal " [")
      (process-children)
      (literal "]")))

  (element edtext
    (process-children)))

;; ============================== INLINES ===============================

(element bibref
  (let ((ref (attribute-string "ref")))
    (make element gi: "A"
	  attributes: (list (list "HREF"
				  (string-append "#"
						 (attribute-string "ref"))))
	  (literal
	   (string-append "["
			  (attribute-string "key"
					    (element-with-id ref))
			  "]")))))

(element emph ($italic-seq$))

(element titleref ($italic-seq$))

(element termref
  (let ((target (element-with-id (attribute-string "def"))))
    (make element gi: "A"
	  attributes: (list (list "HREF" (href-to target)))
	  (process-children))))

(element xtermref ($italic-seq$))

(element specref
  (let* ((id (attribute-string "ref"))
	 (target (element-with-id id)))
    (make element gi: "A"
	  attributes: (list (list "HREF" (href-to target)))
	  (with-mode specref-head-mode
	    (process-node-list
	     (select-elements (children target) "head"))))))

(mode specref-head-mode
  (element head
	   (sosofo-append
	    (make element gi: "B"
		  (literal
		   (cond
		    ((have-ancestor? "body")
		     (REFNUMLABEL "1"))
		    ((have-ancestor? "back")
		     (REFNUMLABEL "A"))
		    (else ""))))
	    (make element gi: "B"
		  (process-children)))))

(element termdef
  (sosofo-append
   (make element gi: "A"
	 attributes: (list (list "NAME" (attribute-string "id")))
	 (empty-sosofo))
   (process-children)))

(element term ($italic-seq$)) ;; should the DTD allow free floating terms?
(element (termdef term) ($italic-seq$))

(element loc ;; currently just inherited properties
  (make element gi: "A"
	attributes: (list (list "HREF" (attribute-string "href"))
			  (list "CLASS" "loc"))
	(process-children)))

(element code
  (let* ((role (attribute-string "role"))
	 (content (make element gi: "CODE"
			attributes: (if role
					(list (list "CLASS" role))
					'()))))
     
    (case role
      (("attrvalue")
       (make element gi: "I" content))
      (("elemname" "featurename")
       (make element gi: "B" content))
      (else content))))

(define (class-attlist #!optional (node (current-node)))
  (list (list "CLASS" (gi node))))

(element foname (make element gi: "P"
		      attributes: '(("CLASS" "foname"))
		      (process-children)))

(element attqual (make element gi: "BLOCKQUOTE" 
		       attributes: '(("CLASS" "attqual"))
		       (process-children)))

(element attdef (make element gi: "P"
		      attributes: '(("CLASS" "attdef"))
		      (process-children)))

(element fosummary (make element gi: "BLOCKQUOTE"
			 attributes: '(("CLASS" "fosummary"))
			 (process-children)))

(element attsummary (make element gi: "BLOCKQUOTE"
			  attributes: '(("CLASS" "attsummary"))
			  (process-children)))

(element kw ($mono-seq$))

(element quote
  (make sequence
    (literal "\"")
    (process-children-trim)
    (literal "\"")))

(element lit
  (make element gi: "SPAN"
	(sosofo-append
	 (literal (string #\U-2018))
	 (make sequence
	   (process-children))
	 (literal (string #\U-2019)))))

(element xnt
  (make element gi: "A"
	attributes: (list (list "HREF" (attribute-string "href")))
	(make element gi: (if (have-ancestor? "prod")
			      "SPAN"
			      "I")
	      (process-children))))

(element nt
  (let* ((def    (attribute-string "def"))
	 (target (if def
		     (element-with-id def)
		     (empty-node-list))))
    (if def
	(make element gi: "A"
	      attributes: (list (list "HREF" (href-to target)))
	      (if (have-ancestor? "prod")
		  (make element gi: "SPAN"
			(process-children))
		  (make element gi: "I"
			(process-children))))
	(if (have-ancestor? "prod")
	    (make element gi: "SPAN"
		  (process-children))
	    (make element gi: "I"
		  (process-children))))))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Code for my-preced and my-follow contributed by Bill Lindsey

;; different from ipreced in that it works on a singleton node list.
;; returns the preceding sibling element or empty node list if
;;  this is the first child.

(define (my-preced snl)
  (let loop ((scanned (empty-node-list))
	     (rest (siblings snl)))
    (cond ((node-list-empty? rest)
	   (empty-node-list))
	  ((node-list=? (node-list-first rest) snl)
	   scanned)
	  (else
	   (loop (node-list-first rest)
		 (node-list-rest rest))))))

(define (my-follow snl)
  (let ((sibs (siblings snl)))
    (let loop ((scanned (node-list-first sibs))
	       (rest (node-list-rest sibs)))
      (cond ((node-list-empty? rest)
	     (empty-node-list))
	    ((node-list=? scanned snl)
	     (node-list-first rest))
	    (else
	     (loop (node-list-first rest)
		   (node-list-rest rest)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; A couple of necessary utilities by Bill Lindsey
;;
;; 1) get the list of sibling elements (assumes origin == parent,
;;                                   in other words we're content)

(define (siblings snl)
  (children (parent snl)))

;; 2) (empty-node-list)
;; this doesn't seem to have been supplied with jade, so we go find
;;  an empty node list at the tail of the "current-node" node-list.

(define (empty-node-list)      
  (let loop ((enl (current-node)))
    (if (node-list-empty? enl)
	enl
	(loop (node-list-rest enl)))))

;; Translation of the HTML table model into HTML ought to be pretty
;; straightforward, yes?

(define (copy-attributes #!optional (nd (current-node)))
  (let loop ((atts (named-node-list-names (attributes nd))))
    (if (null? atts)
        '()
        (let* ((name (car atts))
               (value (attribute-string name nd)))
          (if value
              (cons (list name value)
                    (loop (cdr atts)))
              (loop (cdr atts)))))))

(element htable
  (make element gi: "TABLE"
	attributes: (copy-attributes)
	(process-children)))

(element htbody
  (process-children))

(element tr
  (make element gi: "TR"
	attributes: (copy-attributes)
	(process-children)))

(element th
  (make element gi: "TH"
	attributes: (copy-attributes)
	(process-children)))

(element td
  (make element gi: "TD"
	attributes: (copy-attributes)
	(process-children)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Routines from ISO/IEC 10179

(define (node-list-reduce nl
			  combine
			  init)
  (if (node-list-empty? nl)
      init
      (node-list-reduce (node-list-rest nl)
			combine
			(combine init
				 (node-list-first nl)))))

(define (node-list-filter proc
			  nl)
  (node-list-reduce nl
		    (lambda (result snl)
		      (if (proc snl)
			  (node-list snl
				     result)
			  result))
		    (empty-node-list)))

(define (node-list-property prop nl)
  (node-list-map (lambda (snl)
                   (node-property prop snl default:
 (empty-node-list)))
                 nl))

(define (origin nl)
  (node-list-property 'origin nl))

;; ====================== DECLARATIONS FOR JADE ========================

(declare-flow-object-class element
  "UNREGISTERED::James Clark//Flow Object Class::element")

(declare-flow-object-class empty-element
  "UNREGISTERED::James Clark//Flow Object Class::empty-element")

(declare-flow-object-class document-type
  "UNREGISTERED::James Clark//Flow Object Class::document-type")

(declare-flow-object-class processing-instruction
  "UNREGISTERED::James Clark//Flow Object Class::processing-instruction")

(declare-flow-object-class entity
  "UNREGISTERED::James Clark//Flow Object Class::entity")

(declare-flow-object-class entity-ref
  "UNREGISTERED::James Clark//Flow Object Class::entity-ref")

(declare-flow-object-class formatting-instruction
  "UNREGISTERED::James Clark//Flow Object Class::formatting-instruction")

(define debug
  (external-procedure "UNREGISTERED::James Clark//Procedure::debug"))

(define all-element-number
  (external-procedure "UNREGISTERED::James Clark//Procedure::all-element-number"))

;; Make text that comes from unimplemented tags easy to spot
(default
  (make element gi: "FONT"
	attributes: '(("COLOR" "RED"))
	(literal (gi (current-node)) ": ")
	(process-children)))

(define ($w3c-legal-boilerplate$)
  (make element gi: "p"
      (make element gi: "small"
	    (make element
		  gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.w3.org/Consortium/Legal/ipr-notice.html#Copyright"))
		  (literal "Copyright"))
	    (literal " ")
	    (make formatting-instruction data: "&#38;nbsp;&#38;copy;&#38;nbsp;")
	    (literal " 1998 ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.w3.org"))
		  (literal "W3C"))
	    (literal " (")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.lcs.mit.edu"))
		  (literal "MIT"))
	    (literal ", ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.inria.fr/"))
		  (literal "INRIA"))
	    (literal ", ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.keio.ac.jp/"))
		  (literal "Keio"))
	    (literal "), All Rights Reserved. ")
	    (literal "W3C ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.w3.org/Consortium/Legal/ipr-notice.html#Legal Disclaimer"))
		  (literal "liability"))
	    (literal ", ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.w3.org/Consortium/Legal/ipr-notice.html#W3C Trademarks"))
		  (literal "trademark"))
	    (literal ", ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.w3.org/Consortium/Legal/copyright-documents.html"))
		  (literal "document use"))
	    (literal " and ")
	    (make element gi: "a"
		  attributes: (list (list "HREF"
					  "http://www.w3.org/Consortium/Legal/copyright-software.html"))
		  (literal "software licensing"))
	    (literal " rules apply."))))

(element (p ednote)
	 (sosofo-append
	  (literal " [Ednote: ")
	  (process-children)
	  (literal "]")))

(element (p ednote edtext) (process-children))
